/*------------------------------------------------------------------------------
* postpos.c : post-precessing positioning
*
*          Copyright (C) 2007 by T.TAKASU, All rights reserved.
*
* version : $Revision: 5 $ $Date: 07/05/10 2:36 $
* history : 2007/05/08  1.0 new
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id:$";
static const char prog[]="rtklib ver.1.2";

/* type ----------------------------------------------------------------------*/
typedef struct {            /* analysis results struct */
    gtime_t t;              /* time */
    sol_t sol,solf;         /* float, dgps or single solution, fixed solution */
    int stat,ns;            /* status, number of satellite */
} result_t;

/* external funcitons --------------------------------------------------------*/
extern void printmsg(const char *format, ...);
extern int printprog(int prog);

/* constants/global variables ------------------------------------------------*/
#define ELMINR      15.0    /* elevation mask for reference station (deg) */
#define COMMENTH    "%"     /* comment line indicator */

static opt_t opt;           /* positioning option */
static gtime_t ts={0,0};    /* start time */
static gtime_t te={0,0};    /* end time */
static double tint=0.0;     /* time interval (sec) */
static int dirs=0;          /* solutions (0:forward,1:backward,2:combined) */
static int times=0;         /* time system (0:gpst,1:utc) */
static int timef=0;         /* time form (0:sssss.s,1:yyyy/mm/dd hh:mm:ss.s) */
static int timeu=3;         /* columns of time under decimal point */
static int posf=0;          /* position form (0:lat,lon,hgt,1:x,y,z-ecef,2:nmea) */
static int degf=0;          /* latitude/longitude form (0:ddd.ddd,1:ddd mm ss) */
static char *sep=" ";       /* field separator */
static char corrfile[256]=""; /* correction file */

static FILE *fp;            /* output file pointer */
static obs_t obss;          /* observation data */
static nav_t navs;          /* navigation messages */
static int nepoch=0;        /* number of observation epochs */
static int iobsu=0;         /* current rover observation data record index */
static int iobsr=0;         /* current reference observation data record index */
static int revs=0;          /* analysis direction (0:forward,1:backward) */
static gtime_t ts;          /* start epoch */
static double tspan=1.0;    /* time span (sec) */
static result_t *resf;      /* forward analysis results */
static result_t *resb;      /* backward analysis results */
static int iresf=0;         /* current forward analysis result index */
static int iresb=0;         /* current backward analysis result index */
static corr_t *corrs=NULL;  /* corrections */
static int ncorr=0;         /* number of corrections */
static int icorr=0;         /* current correction index */

/* deg to deg/min/sec --------------------------------------------------------*/
static void todms(double deg, double *dms)
{
    double sgn=1.0;
    if (deg<0.0) {deg=-deg; sgn=-1.0;}
    dms[0]=sgn*floor(deg);
    dms[1]=floor((deg-dms[0])*60.0);
    dms[2]=(deg-dms[0]-dms[1]/60.0)*3600;
}
/* compute reference station position as average of single position ----------*/
static int refpos(double *rb)
{
    int i,j,k,n;
    double rr[3]={0.0},Q[9],dtr;
    
    for (i=n=0;i<obss.n;i=j) {
        for (;i<obss.n;i++) if (obss.data[i].rcv==2) break;
        for (j=i;j<obss.n;j++) {
            if (timediff(obss.data[j].time,obss.data[i].time)>0.0) break;
        }
        if (pntpos(obss.data+i,j-i,&navs,ELMINR*D2R,rr,Q,&dtr)<0) continue;
        for (k=0;k<3;k++) rb[k]+=rr[k]; n++;
    }
    if (n<=0) return -1; else for (k=0;k<3;k++) rb[k]/=n;
    return 0;
}
/* output header -------------------------------------------------------------*/
static void outhead(char **file, int n)
{
    int i,w1,w2;
    char s1[32],s2[32],*s3[]={"single","dgps","kinematic","static"};
    char *s4[]={"C1","L1","L1+L2"},*s5,*s6,*s7[]={"","/backward","/combined"};
    double t1,t2,pos[3],dms1[3],dms2[3];
    gtime_t ts,te;
    
    fprintf(fp,"%s program   : %s\n",COMMENTH,prog);
    fprintf(fp,"%s inputs    : ",COMMENTH);
    for (i=0;i<n;i++) fprintf(fp,"%s%s",file[i],i<n-1?" ":"");
    fprintf(fp,"\n");
    ts=obss.data[0].time;
    te=obss.data[obss.n-1].time;
    time2str(times?gpst2utc(ts):ts,s1,1);
    time2str(times?gpst2utc(te):te,s2,1);
    t1=time2gpst(ts,&w1);
    t2=time2gpst(te,&w2);
    s5=times?"UTC":"GPST";
    s6=posf!=1?" lat/lon/height=WGS84/ellipsoidal,":" x/y/z-ecef=WGS84,";
    fprintf(fp,"%s obs start : %s %s (gpsweek%04d %7.1fs)\n",COMMENTH,s1,s5,w1,t1);
    fprintf(fp,"%s obs end   : %s %s (gpsweek%04d %7.1fs)\n",COMMENTH,s2,s5,w2,t2);
    fprintf(fp,"%s mode/obsv : %s/%s%s%s%s\n",COMMENTH,s3[opt.mode],
            s4[!opt.mode?0:(opt.nf>1?2:1)],opt.modear?"/instantaneous AR":"",
            opt.thresar<=0.0?"/float":"",s7[dirs]);
    fprintf(fp,"%s elev mask : %4.1f deg\n",COMMENTH,opt.elmin*R2D);
    if (opt.mode>=1) {
        fprintf(fp,"%s ref pos   :",COMMENTH);
        if (posf==1) {
            fprintf(fp,"%14.4f%s%14.4f%s%14.4f\n",opt.rb[0],sep,opt.rb[1],sep,opt.rb[2]);
        }
        else {
            ecef2pos(opt.rb,pos);
            if (degf) {
                todms(pos[0]*R2D,dms1); todms(pos[1]*R2D,dms2);
                fprintf(fp,"%3.0f%s%02.0f%s%08.5f%s%4.0f%s%02.0f%s%08.5f%s%10.4f\n",
                        dms1[0],sep,dms1[1],sep,dms1[2],sep,dms2[0],sep,dms2[1],
                        sep,dms2[2],sep,pos[2]);
            }
            else {
                fprintf(fp,"%13.9f%s%14.9f%s%10.4f\n",pos[0]*R2D,sep,pos[1]*R2D,
                        sep,pos[2]);
            }
        }
    }
    fprintf(fp,"%s\n%s (time=%s,%s Q=1:fix,2:float,4:dgps,5:single,ns=# of sats)\n",
            COMMENTH,COMMENTH,s5,s6);
    fprintf(fp,"%s  %-*s",COMMENTH,(timef?17:5)+(timeu<=0?0:timeu+1),"time");
    if (posf==1) {
        fprintf(fp,"%14s %14s %14s %3s %3s %8s %8s %8s\n","x-ecef(m)","y-ecef(m)",
                "z-ecef(m)","Q","ns","sdx(m)","sdy(m)","sdz(m)");
    }
    else if (degf) {
        fprintf(fp,"%16s %16s %10s %3s %3s %8s %8s %8s\n","latitude(d'\")",
                "longitude(d'\")","height(m)","Q","ns","sdn(m)","sde(m)","sdu(m)");
    }
    else {
        fprintf(fp,"%14s %14s %10s %3s %3s %8s %8s %8s\n","latitude(deg)",
                "longitude(deg)","height(m)","Q","ns","sdn(m)","sde(m)","sdu(m)");
    }
}
/* output result in the form of x/y/z-ecef ----------------------------------*/
static void outecef(FILE *fp, const char *s, double *rr, double *Qr, int stat,
                    int ns)
{
    fprintf(fp,"%s%s%14.4f%s%14.4f%s%14.4f%s%3d%s%3d%s%8.4f%s%8.4f%s%8.4f\n",
            s,sep,rr[0],sep,rr[1],sep,rr[2],sep,stat,sep,ns,sep,sqrt(Qr[0]),sep,
            sqrt(Qr[4]),sep,sqrt(Qr[8]));
}
/* output result in the form of lat/lon/height -------------------------------*/
static void outpos(FILE *fp, const char *s, double *rr, double *Qr, int stat,
                   int ns)
{
    double pos[3],dms1[3],dms2[3],Qe[9];
    
    ecef2pos(rr,pos); covenu(pos,Qr,Qe);
    if (degf) {
        todms(pos[0]*R2D,dms1); todms(pos[1]*R2D,dms2);
        fprintf(fp,"%s%s%4.0f%s%02.0f%s%08.5f%s%4.0f%s%02.0f%s%08.5f",s,sep,
                dms1[0],sep,dms1[1],sep,dms1[2],sep,dms2[0],sep,dms2[1],sep,
                dms2[2]);
    }
    else fprintf(fp,"%s%s%14.9f%s%14.9f",s,sep,pos[0]*R2D,sep,pos[1]*R2D);
    fprintf(fp,"%s%10.4f%s%3d%s%3d%s%8.4f%s%8.4f%s%8.4f\n",sep,pos[2],sep,
            stat,sep,ns,sep,sqrt(Qe[4]),sep,sqrt(Qe[0]),sep,sqrt(Qe[8]));
}
/* output result in the form of NMEA-0183 GGA sentence -----------------------*/
static void outnmea(FILE *fp, gtime_t t, double *rr, int stat, int ns)
{
    char buff[256],*p,sum;
    double h,ep[6],pos[3],dms1[3],dms2[3];
    int q[]={0,5,4,0,2,1}; /* quality: 1:sigle,2:dgps,4:float,5:fix,0:other */
    
    time2epoch(gpst2utc(t),ep);
    ecef2pos(rr,pos);
    h=geoidh(pos);
    todms(fabs(pos[0])*R2D,dms1); todms(fabs(pos[1])*R2D,dms2);
    sprintf(buff,"$GPGGA,%02.0f%02.0f%05.2f,%02.0f%010.7f,%s,%03.0f%010.7f,%s,%d,%02d,,%.3f,M,%.3f,M,,,",
            ep[3],ep[4],ep[5],dms1[0],dms1[1]+dms1[2]/60.0,pos[0]>=0?"N":"S",
            dms2[0],dms2[1]+dms2[2]/60.0,pos[1]>=0?"E":"W",q[stat],ns,
            pos[2]-h,h);
    for (p=buff+1,sum=0;*p;p++) sum^=*p; /* check-sum */
    fprintf(fp,"%s*%02X%c%c",buff,sum,0x0D,0x0A);
}
/* search next observation data index ----------------------------------------*/
static int nextobsf(const obs_t *obs, int *i, int rcv)
{
    int n;
    double tt;
    for (;*i<obs->n;(*i)++) if (obs->data[*i].rcv==rcv) break;
    for (n=0;*i+n<obs->n;n++) {
        tt=timediff(obs->data[*i+n].time,obs->data[*i].time);
        if (obs->data[*i+n].rcv!=rcv||tt>MAXDT) break;
    }
    return n;
}
static int nextobsb(const obs_t *obs, int *i, int rcv)
{
    int n;
    double tt;
    for (;*i>=0;(*i)--) if (obs->data[*i].rcv==rcv) break;
    for (n=0;*i-n>=0;n++) {
        tt=timediff(obs->data[*i-n].time,obs->data[*i].time);
        if (obs->data[*i-n].rcv!=rcv||tt<-MAXDT) break;
    }
    return n;
}
/* input observation data/navigation messages for an epoch -------------------*/
static int input(obsd_t **obs, nav_t **nav, corr_t **corr)
{
    static obsd_t obse[MAXOBS];
    int i,nu,nr,n=0;
    *corr=NULL;
    
    if (0<=iobsu&&iobsu<obss.n&&tspan>0.0) {
        if (printprog((int)(timediff(obss.data[iobsu].time,ts)/tspan*100.0+0.5))) {
            printmsg("aborted"); return 0;
        }
    }
    if (!revs) { /* forward */
        if ((nu=nextobsf(&obss,&iobsu,1))<=0) return 0;
        if (opt.intpref) {
            for (;(nr=nextobsf(&obss,&iobsr,2))>0;iobsr+=nr)
                if (timediff(obss.data[iobsr].time,obss.data[iobsu].time)>-MAXDT) break;
        }
        else {
            for (i=iobsr;(nr=nextobsf(&obss,&i,2))>0;iobsr=i,i+=nr)
                if (timediff(obss.data[i].time,obss.data[iobsu].time)>MAXDT) break;
        }
        nr=nextobsf(&obss,&iobsr,2);
        for (i=0;i<nu;i++) obse[n++]=obss.data[iobsu+i];
        for (i=0;i<nr;i++) obse[n++]=obss.data[iobsr+i];
        if (corrs) {
            while (icorr<ncorr&&timediff(corrs[icorr].time,obss.data[iobsu].time)<-0.5)
                icorr++;
            if (icorr<ncorr&&timediff(corrs[icorr].time,obss.data[iobsu].time)<0.5)
                *corr=corrs+icorr;
        }
        iobsu+=nu;
    }
    else { /* backward */
        if ((nu=nextobsb(&obss,&iobsu,1))<=0) return 0;
        if (opt.intpref) {
            for (;(nr=nextobsb(&obss,&iobsr,2))>0;iobsr-=nr)
                if (timediff(obss.data[iobsr].time,obss.data[iobsu].time)<MAXDT) break;
        }
        else {
        for (i=iobsr;(nr=nextobsb(&obss,&i,2))>0;iobsr=i,i-=nr)
            if (timediff(obss.data[i].time,obss.data[iobsu].time)<-MAXDT) break;
        }
        nr=nextobsb(&obss,&iobsr,2);
        for (i=0;i<nu;i++) obse[n++]=obss.data[iobsu-nu+1+i];
        for (i=0;i<nr;i++) obse[n++]=obss.data[iobsr-nr+1+i];
        if (corrs) {
            while (icorr>=0&&timediff(corrs[icorr].time,obss.data[iobsu].time)>0.5)
                icorr--;
            if (icorr>=0&&timediff(corrs[icorr].time,obss.data[iobsu].time)>-0.5)
                *corr=corrs+icorr;
        }
        iobsu-=nu;
    }
    *obs=obse;
    *nav=&navs;
    return n;
}
/* output result -------------------------------------------------------------*/
static void output(gtime_t t, sol_t sol, sol_t solf, int stat, int ns)
{
    int week;
    char s[64];
    double gpst,*rr,*Qr;
    
    rr=stat==1?solf.rr:sol.rr; Qr=stat==1?solf.Qr:sol.Qr;
    if (times) t=gpst2utc(t);
    if (timef) time2str(t,s,timeu);
    else {
        gpst=time2gpst(t,&week);
        if (86400*7-gpst<0.5/pow(10.0,timeu)) gpst=0.0;
        sprintf(s,"%*.*f",7+(timeu<=0?0:timeu+1),timeu,gpst);
    }
    switch (posf) {
        case 1:  outecef(fp,s,rr,Qr,stat,ns); break;
        case 2:  outnmea(fp,t,rr,stat,ns);    break;
        default: outpos(fp,s,rr,Qr,stat,ns);  break;
    }
}
/* save result ---------------------------------------------------------------*/
static void saveres(gtime_t t, sol_t sol, sol_t solf, int stat, int ns)
{
    result_t *p;
    if ((revs?iresb:iresf)>=nepoch) return;
    p=revs?resb+iresb++:resf+iresf++;
    p->t=t; p->sol=sol; p->solf=solf; p->stat=stat; p->ns=ns;
}
/* read correction file ------------------------------------------------------*/
static int readcorr(const char *file)
{
    FILE *fp;
    char buff[1024],*p;
    int i,n=0;
    double ep[6];
    
    corrs=NULL; ncorr=0;
    if (file[0]=='\0') return 0;
    
    if (!(fp=fopen(file,"r"))) return -1;
    
    while (fgets(buff,sizeof(buff),fp)) {
        if (ncorr>=n) {
            n+=4096;
            corrs=(corr_t *)realloc(corrs,sizeof(corr_t)*n);
        }
        if (sscanf(buff,"%lf/%lf/%lf %lf:%lf:%lf",ep,ep+1,ep+2,ep+3,ep+4,ep+5)<6)
            continue;
        corrs[ncorr].time=epoch2time(ep);
        for (i=0;i<MAXSAT;i++) corrs[ncorr].dtc[i]=0.0;
        for (i=0,p=strtok(buff+20," ");i<MAXSAT&&p;p=strtok(NULL," ")) {
            corrs[ncorr].dtc[i++]=atof(p)*1E-3;
        }
        ncorr++;
    }
    fclose(fp);
    return 0;
}
/* smoothed solution ---------------------------------------------------------*/
static int smoothsol(sol_t solf, sol_t solb, sol_t *sols)
{
    int i;
    double rr[3];
    
    /* Qr=(Qf^-1+Qb^-1)^-1,rr=Qr*(Qf^-1*rf+Qb^-1*rb) */
    if (matinv(solf.Qr,3)||matinv(solb.Qr,3)) return -1;
    for (i=0;i<9;i++) sols->Qr[i]=solf.Qr[i]+solb.Qr[i];
    if (matinv(sols->Qr,3)) return -1;
    matmul("NN",3,1,3,1.0,solf.Qr,solf.rr,0.0,rr);
    matmul("NN",3,1,3,1.0,solb.Qr,solb.rr,1.0,rr);
    matmul("NN",3,1,3,1.0,sols->Qr,rr,0.0,sols->rr);
    return 0;
}
/* combine forward/backward solutions and output results ---------------------*/
static void combres(void)
{
    int i,j,k,stat;
    double tt,dr;
    sol_t sol,solf;
    
    for (i=0,j=iresb-1;i<iresf&&j>=0;i++,j--) {
        tt=timediff(resf[i].t,resb[j].t);
        if (tt<-MAXDT) j++; else if (tt>MAXDT) i--;
        if (tt<-MAXDT||resf[i].stat<resb[j].stat) {
            output(resf[i].t,resf[i].sol,resf[i].solf,resf[i].stat,resf[i].ns);
            continue;
        }
        else if (tt>MAXDT||resf[i].stat>resb[j].stat) {
            output(resb[j].t,resb[j].sol,resb[j].solf,resb[j].stat,resb[j].ns);
            continue;
        }
        /* validation of forward/backward fixed solutions */
        if ((stat=resf[i].stat)==1) {
            for (k=0;k<3;k++) {
                dr=resf[i].solf.rr[k]-resb[j].solf.rr[k];
                if (dr*dr>16.0*(resf[i].solf.Qr[k+k*3]+resb[j].solf.Qr[k+k*3]))
                    stat=2; /* degrade to float solution */
            }
        }
        if (stat==1) { /* fixed */
            if (smoothsol(resf[i].solf,resb[j].solf,&solf)) continue;
        }
        else { /* float,dgps or single */
            if (smoothsol(resf[i].sol,resb[j].sol,&sol)) continue;
        }
        output(timeadd(resf[i].t,-tt/2.0),sol,solf,stat,resf[i].ns);
    }
}
/* execute positioning -------------------------------------------------------*/
static int execpos(char **infile, int n, char *outfile)
{
    int i;
    
    /* read rinex obs/nav files */
    if ((nepoch=readrnxt(infile,n,ts,te,tint,&obss,&navs))<=0) {
        printmsg("error : no observation data");
        return -3;
    }
    if (navs.n<=0) {
        printmsg("error : no navigation message");
        return -4;
    }
    /* compute reference station position */
    if (opt.mode>=1&&norm(opt.rb,3)<=0.0&&refpos(opt.rb)) {
        printmsg("error : no reference position");
        return -5;
    }
    for (i=0,ts.time=0;i<obss.n;i++) {
        if (obss.data[i].rcv!=1) continue;
        if (ts.time==0) ts=obss.data[i].time;
        tspan=timediff(obss.data[i].time,ts);
    }
    /* read correction file */
    if (readcorr(corrfile)) {
        printmsg("error : correction file open error : %s",corrfile);
        return -6;
    }
    if (!*outfile) fp=stdout;
    else if (!(fp=fopen(outfile,"w"))) {
        printmsg("error : output file open error : %s",outfile);
        return -6;
    }
    /* output header */
    if (posf<2) outhead(infile,n);
    
    iobsu=iobsr=revs=0;
    if (dirs<2) { /* forward/backward solution */
        if (dirs) {revs=1; iobsu=iobsr=obss.n-1; icorr=ncorr-1;}
        rtkpos(input,output,&opt);
    }
    else { /* combined solution */
        resf=(result_t *)malloc(sizeof(result_t)*nepoch);
        resb=(result_t *)malloc(sizeof(result_t)*nepoch);
        if (resf&&resb) {
            iresf=iresb=icorr=0;
            rtkpos(input,saveres,&opt);
            revs=1; iobsu=iobsr=obss.n-1; icorr=ncorr-1;
            rtkpos(input,saveres,&opt);
            
            /* combine forward/backward solutions and output results */
            combres();
        }
        else printmsg("error : memory allocation error");
        free(resf); free(resb);
    }
    fclose(fp);
    return 0;
}
/* post positioning -------------------------------------------------------------
* post-processing positioning
* args   : gtime_t ts,te    I   estimation start/end time
*          double tint      I   estimation interval (s) (0:all)
*          opt_t  *opt      I   processing options
*          char   *sep      I   output field separator
*          int    timeu     I   digits under decimal point
*          int    dirs      I   proc direction (0:forward,1:backward,2:combined)
*          int    times     I   time system
*          int    timef     I   time format
*          int    posf      I   solution format
*          int    degf      I   latitude/longitude format
*          char   *infile[] I   input files
*          int    nfile     I   number of input files
*          char   *outfile  I   output file
*          char   *corrfile I   correction file
* return : status (0:ok,0>:error)
*-----------------------------------------------------------------------------*/
extern int postpos(gtime_t ts_, gtime_t te_, double tint_, const opt_t *opt_,
                   char *sep_, int timeu_, int dirs_, int times_, int timef_,
                   int posf_, int degf_, char **infile, int nfile, char *outfile,
                   char *corrfile_)
{
    ts=ts_; te=te_; tint=tint_; opt=*opt_; sep=sep_; timeu=timeu_; dirs =dirs_;
    times=times_; timef=timef_; posf=posf_; degf =degf_;
    strcpy(corrfile,corrfile_);
    return execpos(infile,nfile,outfile);
}
