/*------------------------------------------------------------------------------
* rtkpos.c : realtime kinematic positioning
*
*          Copyright (C) 2007 by T.TAKASU, All rights reserved.
*
* version : $Revision: 7 $ $Date: 07/05/10 2:37 $
* history : 2007/01/12 1.0 new
*           2007/03/13 1.1 add slip detection by LLI flag
*           2007/04/18 1.2 add antenna pcv correction
*                          change rtkpos argin
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id: rtkpos.c,v 1.12 2007/03/20 08:49:29 TTAKA Exp TTAKA $";

/* constants/macros ----------------------------------------------------------*/

#define VARB    100.0      /* initial phase-bias variance (cycle^2) */
#define VARI    0.001      /* initial ionosphere delay variance (m^2/10km) */
#define VART    0.001      /* initial troposphere delay variance (m^2/10km) */
#define MAXSTD  1.0        /* max standard deviation of fix solution (m) */

#define NP(opt)     (opt->mode<4?3:0) /* number of pos parameters */
#define II(s,opt)   (NP(opt)+(s)-1)   /* state index of ionospheric delay */
#define IT(opt)     (NP(opt)+(opt->estiono?MAXSAT:0)) /* state index of tropos */
#define NR(opt)     (IT(opt)+(opt->esttrop?1:0)) /* number of real prms */
#define IB(s,f,opt) (NR(opt)+MAXSAT*(f)+(s)-1) /* state index of phase-bias */

const static double fion[][2]={  /* ionospheric delay factor relative to P1 */
    {-1.0,1.0},
    {-FREQ1*FREQ1/FREQ2/FREQ2,FREQ1*FREQ1/FREQ2/FREQ2},
    {-FREQ1*FREQ1/FREQ5/FREQ5,FREQ1*FREQ1/FREQ5/FREQ5}
};
/* types --------------------------------------------------------------------*/
typedef struct {                /* ambiguity control struct */
    gtime_t t;                  /* time */
    int lock[MAXSAT][NFREQ];    /* lock counter of phase */
    int out[MAXSAT][NFREQ];     /* obs outage counter for phase */
    double gf[MAXSAT];          /* geometry-free phase (m) */
} ambctr_t;

/* single-differenced observable (f=0:L1,1:L2,2:C1,3:P2) ---------------------*/
static double sdobs(const obsd_t *obs, int i, int j, int f)
{
    double pi=f<NFREQ?lam[f]*obs[i].L[f]:obs[i].P[f-NFREQ];
    double pj=f<NFREQ?lam[f]*obs[j].L[f]:obs[j].P[f-NFREQ];
    return pi==0.0||pj==0.0?0.0:pi-pj;
}
/* single-differenced geometry-free linear combination of phase --------------*/
static double gfobs(const obsd_t *obs, int i, int j)
{
    double pi=sdobs(obs,i,j,0),pj=sdobs(obs,i,j,1);
    return pi==0.0||pj==0.0?0.0:pi-pj;
}
/* single-differenced measurement error variance (f=0:L1,1:L2,2:C1,3:P2) -----*/
static double varerr(double el, double bl, double dt, int f, const opt_t *opt)
{
    double a=opt->err[1],b=opt->err[2]/sin(el),c=opt->err[3]*bl/1E4;
    double d=CLIGHT*opt->sclkstab*dt;
    return 2.0*(a*a+b*b+c*c)*(f<NFREQ?1.0:opt->err[0]*opt->err[0])+d*d;
}
/* baseline length -----------------------------------------------------------*/
static double baseline(const double *ru, const double *rb)
{
    int i;
    double dr[3];
    for (i=0;i<3;i++) dr[i]=ru[i]-rb[i];
    return norm(dr,3);
}
/* select common satellites between rover and reference station --------------*/
static int selsats(const obsd_t *obs, const double *azel, int nu, int nr,
                   double elmin, int *sat, int *iu, int *ir)
{
    int i,j,k=0;
    for (i=0,j=nu;i<nu&&j<nu+nr;i++,j++) {
        if      (obs[i].sat<obs[j].sat) j--;
        else if (obs[i].sat>obs[j].sat) i--;
        else if (azel[1+i*2]>=elmin) {sat[k]=obs[i].sat; iu[k]=i; ir[k++]=j;}
    }
    return k;
}
/* initialize state and covariance -------------------------------------------*/
static void initx(double *x, double *P, int nx, double xi, double var, int i)
{
    int j;
    x[i]=xi; for (j=0;j<nx;j++) P[i+j*nx]=P[j+i*nx]=i==j?var:0.0;
}
/* temporal update of states -------------------------------------------------*/
static void udstate(double *x, double *P, int nx, const obsd_t *obs,
                    const int *sat, const int *iu, const int *ir, int ns,
                    double bl, ambctr_t *ambc, const opt_t *opt)
{
    int i,j,k,f,reset,slip[MAXSAT]={0};
    double dt,g0,g1,cp,pr,bias[MAXSAT]={0.0},offset;
    gtime_t t=obs[0].time;
    
    dt=ambc->t.time==0?0.0:timediff(t,ambc->t); ambc->t=t;
    
    for (i=0;i<ns;i++) {
        /* detect slip by LLI flag or geometry-free phase-jump */
        slip[i]=obs[iu[i]].LLI[0]||obs[ir[i]].LLI[0];
        if ((g1=gfobs(obs,iu[i],ir[i]))==0.0) continue;
        g0=ambc->gf[sat[i]-1]; ambc->gf[sat[i]-1]=g1;
        slip[i]|=g0!=0.0&&fabs(g1-g0)>opt->thresslip;
    }
    for (f=0;f<opt->nf;f++) {
        /* reset phase-bias if instantaneous AR or expire obs outage counter */
        for (i=1;i<=MAXSAT;i++) {
            reset=++ambc->out[i-1][f]>opt->maxout;
            if ((opt->modear==1||reset)&&x[IB(i,f,opt)]!=0.0) {
                initx(x,P,nx,0.0,0.0,IB(i,f,opt));
            }
            if (opt->modear!=1&&reset) ambc->lock[i-1][f]=-opt->minlock;
        }
        /* reset phase-bias if slip */
        for (i=0;i<ns;i++) {
            k=IB(sat[i],f,opt); P[k+k*nx]+=opt->prn[0]*opt->prn[0]*fabs(dt);
            if (opt->modear==1||!slip[i]) continue;
            x[k]=0.0; ambc->lock[sat[i]-1][f]=-opt->minlock;
        }
        for (i=j=0,offset=0.0;i<ns;i++) {
            /* approx single-differenced phase-bias by phase - pseudorange */
            cp=sdobs(obs,iu[i],ir[i],f); pr=sdobs(obs,iu[i],ir[i],NFREQ);
            if (cp!=0.0&&pr!=0.0) bias[i]=(cp-pr)/lam[f]; 
            if (x[IB(sat[i],f,opt)]!=0.0) {offset+=bias[i]-x[IB(sat[i],f,opt)]; j++;}
        }
        /* correct phase-bias offset caused by receiver clock-jump */
        if (j>0) {
            for (i=1;i<=MAXSAT;i++) if (x[IB(i,f,opt)]!=0.0) x[IB(i,f,opt)]+=offset/j;
        }
        /* set initial phase-bias */
        for (i=0;i<ns;i++) {
            if (bias[i]==0.0||x[IB(sat[i],f,opt)]!=0.0) continue;
            initx(x,P,nx,bias[i],VARB,IB(sat[i],f,opt));
        }
    }
    if (opt->estiono) { /* reset ionosphere parameter */
        for (i=1;i<=MAXSAT;i++) {
            if (x[II(i,opt)]!=0.0&&ambc->out[i-1][0]>10&&ambc->out[i-1][1]>10)
                x[II(i,opt)]=0.0;
        }
        for (i=0;i<ns;i++) {
            k=II(sat[i],opt);
            if (x[k]==0.0) initx(x,P,nx,1E-6,VARI*bl/1E4,k);
            else P[k+k*nx]+=opt->prn[1]*opt->prn[1]*fabs(dt)*bl/1E4;
        }
    }
    if (opt->esttrop) { /* reset tropospheric parameter */
        k=IT(opt);
        if (x[k]==0.0) initx(x,P,nx,1E-6,VART*bl/1E4,k);
        else P[k+k*nx]+=opt->prn[2]*opt->prn[2]*fabs(dt)*bl/1E4;
    }
}
/* undifferenced phase/code residuals ----------------------------------------*/
static void zdres(gtime_t time, const obsd_t *obs, int n, const nav_t *nav,
                  const double *rr, const opt_t *opt, const pcv_t *pcv,
                  double res[][NFREQ*2], double *e, double *azel)
{
    int i,j;
    double pr,pos[3],dr[NFREQ]={0},rs[3*MAXOBS],dts[MAXOBS];
    
    satpos(time,obs,n,nav,rs,dts); /* satellite position/clock-bias */
    ecef2pos(rr,pos);
    
    for (i=0;i<n;i++) {
        if ((pr=geodist(rs+i*3,rr,e+i*3))<=0.0) continue;
        satazel(pos,e+i*3,azel+i*2);
        if (azel[1+i*2]<opt->elmin) continue;
        
        /* satellite clock-bias+tropospheric delay */
        pr+=-CLIGHT*dts[i]+tropmodel(pos,azel+i*2,opt->esttrop?0.0:0.5);
        
        /* antenna phase center */
        antmodel(pcv,azel+i*2,dr);
        
        /* observable - pseudorange model */
        for (j=0;j<NFREQ;j++) {
            if (obs[i].L[j]!=0.0) res[i][j]=lam[j]*obs[i].L[j]-pr-dr[j];
            if (obs[i].P[j]!=0.0) res[i][j+NFREQ]=obs[i].P[j]-pr-dr[j];
        }
    }
}

/* double-differenced phase/code residuals -----------------------------------*/
static int ddres(gtime_t time, double dt, double *x, int nx, const int *sat,
                 double res[][NFREQ*2], double *e, double *azel, const int *iu,
                 const int *ir, int ns, double bl, const opt_t *opt, double *v,
                 double *H, double *var, int *n, int *vs)
{
    int i,j,k,f,nv=0;
    double pos[3],didx=0.0,dtdxi,dtdxj;
    
    for (f=opt->mode>=2?0:NFREQ;f<NFREQ*2;f+=NFREQ-opt->nf+1) {
        
        /* make double-difference */
        for (i=j=0;j<ns;j++) {
            if (res[iu[j]][f]==0.0||res[ir[j]][f]==0.0) continue;
            if (azel[1+iu[j]*2]>=azel[1+iu[i]*2]) i=j;
        }
        for (j=0;j<ns;j++) {
            if (i==j||res[iu[j]][f]==0.0||res[ir[j]][f]==0.0) continue;
              
            /* double-differenced residual */
            v[nv]=(res[iu[i]][f]-res[ir[i]][f])-(res[iu[j]][f]-res[ir[j]][f]);
            
            if (opt->estiono) {
                didx=fion[f%NFREQ][f<NFREQ?0:1];
                v[nv]-=didx*x[II(sat[i],opt)]-didx*x[II(sat[j],opt)];
            }
            if (opt->esttrop) {
                ecef2pos(x,pos);
                tropmapf(time,pos,azel+iu[i]*2,&dtdxi);
                tropmapf(time,pos,azel+iu[j]*2,&dtdxj);
                v[nv]-=(dtdxi-dtdxj)*x[IT(opt)];
            }
            if (f<NFREQ) {
                v[nv]-=lam[f]*(x[IB(sat[i],f,opt)]-x[IB(sat[j],f,opt)]);
            }
            if (fabs(v[nv])>opt->maxinno) continue;
            
            /* partial derivatives */
            if (opt->mode<4) {
                for (k=0;k<3;k++) H[k+nv*nx]=-e[k+iu[i]*3]+e[k+iu[j]*3];
            }
            if (opt->estiono) {
                H[II(sat[i],opt)+nv*nx]= didx;
                H[II(sat[j],opt)+nv*nx]=-didx;
            }
            if (opt->esttrop) H[IT(opt)+nv*nx]=dtdxi-dtdxj;
            if (f<NFREQ) {
                H[IB(sat[i],f,opt)+nv*nx]= lam[f];
                H[IB(sat[j],f,opt)+nv*nx]=-lam[f];
                vs[f+i*NFREQ]=vs[f+j*NFREQ]=1;
            }
            /* measurement error variance */
            var[  nv*2]=varerr(azel[1+iu[i]*2],bl,dt,f,opt);
            var[1+nv*2]=varerr(azel[1+iu[j]*2],bl,dt,f,opt);
            nv++; n[f]++;
        }
    }
    return nv; /* return number of v */
}
/* interpolate reference observables -----------------------------------------*/
static double interpr(gtime_t time, const obsd_t *obs, int n, const nav_t *nav,
                      const double *rb, const opt_t *opt, double res[][NFREQ*2])
{
    static obsd_t obsbuf[MAXOBS]; /* observation data buffer */
    static int nbuf;
    int i,j,k;
    double resbuf[MAXOBS][NFREQ*2]={{0}},e[3*MAXOBS],azel[2*MAXOBS];
    double dt=timediff(time,obs[0].time),dtb=timediff(time,obsbuf[0].time);
    
    if (fabs(dt)<MAXDT) {
        nbuf=n; for (i=0;i<n;i++) obsbuf[i]=obs[i];
    }
    else if (fabs(dtb)<opt->maxtdiff*2.0&&dt!=dtb) { /* interpolate */
        zdres(time,obsbuf,nbuf,nav,rb,opt,&opt->pcvr,resbuf,e,azel);
        for (i=0;i<n;i++) {
            for (j=0;j<nbuf;j++) if (obsbuf[j].sat==obs[i].sat) break;
            if (j>=nbuf) continue;
            for (k=0;k<NFREQ*2;k++) {
                if (res[i][k]==0.0||resbuf[j][k]==0.0) continue;
                res[i][k]=(dtb*res[i][k]-dt*resbuf[j][k])/(dtb-dt);
            }
        }
        if (fabs(dtb)<fabs(dt)) dt=dtb;
    }
    return dt;
}
/* double-differenced measurement error covariance ---------------------------*/
static void coverr(int nv, const int *n, const double *var, const opt_t *opt,
                   double *R)
{
    int i,j,k,f;
    for (f=i=0;f<NFREQ*2;f+=NFREQ-opt->nf+1,i=j) {
        for (j=i;j<i+n[f];j++) {
            for (k=j;k<i+n[f];k++) R[j+k*nv]=R[k+j*nv]=var[j*2];
            R[j+j*nv]+=var[1+j*2];
        }
    }
}
/* relative positioning ------------------------------------------------------*/
static int relpos(double *x, double *P, int nx, const obsd_t *obs, int nu,
                  int nr, const nav_t *nav, const corr_t *corr, ambctr_t *ambc,
                  const double *rb, const opt_t *opt)
{
    int i,j,f,ns,ny,nv,info=-1;
    int n[NFREQ*2]={0},sat[MAXSAT],iu[MAXSAT],ir[MAXSAT],vs[NFREQ*MAXSAT]={0};
    double dt,s,*v,*H,*var,*R,*xp,*Pp,bl=baseline(x,rb);
    double res[MAXOBS][NFREQ*2]={{0}},e[3*MAXOBS],azel[2*MAXOBS]={0};
    gtime_t time=obs[0].time;
    
    if (fabs(dt=timediff(time,obs[nu].time))>opt->maxtdiff) return -1;
    
    /* undifferenced residuals */
    zdres(time,obs,nu,nav,x,opt,&opt->pcvu,res,e,azel);
    zdres(time,obs+nu,nr,nav,rb,opt,&opt->pcvr,res+nu,e+nu*3,azel+nu*2);
    
    /* interpolate reference observables */
    if (opt->intpref) dt=interpr(time,obs+nu,nr,nav,rb,opt,res+nu);
    
    /* satellite clock correction */
    if (corr) {
        for (i=0;i<nr;i++) for (j=0;j<NFREQ*2;j++) {
            res[nu+i][j]+=corr->dtc[obs[nu+i].sat-1];
        }
    }
    /* select common satellites */
    if ((ns=selsats(obs,azel,nu,nr,opt->elmin,sat,iu,ir))<=0) return -1;
    
    /* temporal update of states */
    udstate(x,P,nx,obs,sat,iu,ir,ns,baseline(x,rb),ambc,opt);
    
    ny=ns*opt->nf*NFREQ;
    xp=mat(nx,1); Pp=mat(nx,nx); v=mat(ny,1); H=zeros(nx,ny); var=mat(2,ny);
    
    /* double-differenced residuals (innovation) */
    nv=ddres(time,dt,x,nx,sat,res,e,azel,iu,ir,ns,bl,opt,v,H,var,n,vs);
    
    if ((R=zeros(nv,nv))) {
        /* measurement error covairance */
        coverr(nv,n,var,opt,R);
        
        /* kalman filter state update */
        info=filter(x,P,H,v,R,nx,nv,xp,Pp);
        
        /* postfit residual test */
        zdres(time,obs,nu,nav,xp,opt,&opt->pcvu,res,e,azel);
        nv=ddres(time,dt,xp,nx,sat,res,e,azel,iu,ir,ns,bl,opt,v,H,var,n,vs);
        for (i=0,s=0.0;i<nv;i++) s+=v[i]*v[i]/(var[i*2]+var[1+i*2]);
        if (nv>NR(opt)&&s>chisqr[nv-NR(opt)-1]) info=-1;
        if (!info) {
            /* update ambiguity control struct */
            for (f=0;f<opt->nf;f++) {
                for (i=0;i<ns;i++) {
                    if (!vs[f+i*NFREQ]) continue;
                    ambc->lock[sat[i]-1][f]++; ambc->out[sat[i]-1][f]=0;
                }
            }
            memcpy(x,xp,sizeof(double)*nx); memcpy(P,Pp,sizeof(double)*nx*nx);
        }
    }
    free(xp); free(Pp); free(v); free(H); free(var); free(R);
    return info;
}
/* resolve integer ambiguity -------------------------------------------------*/
static int resamb(double *x, double *P, int nx, const opt_t *opt,
                  ambctr_t *ambc, double *rr, double *Qr)
{
    int i,j,k,f,ny,nr=NR(opt),nb=0,info;
    double *D,*DP,*y,*Qy,*b,*Qb,*Qrb,*QQ,s[2];
    
    /* single to double-difference transformation matrix (D') */
    D=zeros(nx,nx); for (i=0;i<nr;i++) D[i+i*nx]=1.0;
    for (f=0,k=nr;f<opt->nf;f++,k+=MAXSAT) {
        for (i=k  ;i<k+MAXSAT;i++) if (x[i]!=0.0&&ambc->lock[i-k][f]>0) break;
        for (j=i+1;j<k+MAXSAT;j++) if (x[j]!=0.0&&ambc->lock[j-k][f]>0) {
            D[i+(nr+nb)*nx]=1.0; D[j+(nr+nb++)*nx]=-1.0;
        }
    }
    if (nb<=0) {free(D); return -1;} /* no valid double-difference */
    
    /* transform single to double-differenced phase-bias (y=D'*x, Qy=D'*P*D) */
    ny=nr+nb; y=mat(ny,1); Qy=mat(ny,ny); DP=mat(ny,nx);
    matmul("TN",ny,1,nx,1.0,D,x,0.0,y);
    matmul("TN",ny,nx,nx,1.0,D,P,0.0,DP);
    matmul("NN",ny,ny,nx,1.0,DP,D,0.0,Qy);
    free(D); free(DP);
    
    /* phase-bias covariance (Qb) and real-parameters to bias covariance (Qrb) */
    b=mat(nb,2); Qb=mat(nb,nb); Qrb=mat(nr,nb); QQ=mat(nr,nb);
    for (i=0;i<nb;i++) for (j=0;j<nb;j++) Qb[i+j*nb]=Qy[nr+i+(nr+j)*ny];
    for (i=0;i<nr;i++) for (j=0;j<nb;j++) Qrb[i+j*nr]=Qy[i+(nr+j)*ny];
    
    /* lambda/mlambda integer least-square estimation */
    if (!(info=lambda(nb,2,y+nr,Qb,b,s))) {
        /* validation by popular ratio-test */
        if (s[1]/s[0]>=opt->thresar) {
            
            /* transform float to fixed solution (rr=rr-Qrb*Qb\(b0-b)) */
            for (i=0;i<nr;i++) {
                rr[i]=x[i]; for (j=0;j<nr;j++) Qr[i+j*nr]=P[i+j*nx];
            }
            for (i=0;i<nb;i++) y[nr+i]-=b[i];
            if (!(info=matinv(Qb,nb))) {
                matmul("NN",nb,1,nb,1.0,Qb,y+nr,0.0,b);
                matmul("NN",nr,1,nb,-1.0,Qrb,b,1.0,rr);
                
                /* covariance (Qr=Qr-Qrb*Qb^-1*Qrb') */
                matmul("NN",nr,nb,nb,1.0,Qrb,Qb,0.0,QQ);
                matmul("NT",nr,nr,nb,-1.0,QQ,Qrb,1.0,Qr);
            }
        }
        else info=-1; /* validation failed */
    }
    free(y); free(Qy); free(b); free(Qb); free(Qrb); free(QQ);
    return info; /* 0:ok */
}
/* relative positioning --------------------------------------------------------
* input observation data and navigation message, compute rover position by 
* relative positioning and output results. if callback input() return 0, exit
* the function.
* args   : infunc_t input   I   input callback function
*          outfunc_t output I   output callback function
*          opt_t  *opt      I   positioning options (see rtklib.h)
* return : none
* notes  : input and output callback shall comply the following prototypes.
*
*          int input(obsd_t *obs, nav_t *nav);
*            obsd_t *obs    I   observation data records for an epoch
*                               obs[i].rcv=1:rover,2:reference
*                               sorted by receiver and satellte
*            nav_t  *nav    I   navigation messages
*            return             number of observation data records
*
*          void output(gtime_t t, sol_t sol, sol_t solf, int stat, int ns);
*            gtime_t t      I   time
*            sol_t sol,solf I   float, dgps or single solution, fixed solution
*            int    stat    I   status (1:fix,2:float,4:dgps,5:single)
*            int    ns      I   number of valid satellites
*-----------------------------------------------------------------------------*/
extern void rtkpos(infunc_t input, outfunc_t output, const opt_t *opt)
{
    int i,j,nx,ny,nu,nr,ns,stat;
    double *x,*P,*xf,*Qf,Qb[9],dtr[2],var;
    ambctr_t ambc={{0}};
    obsd_t *obs;
    nav_t *nav;
    corr_t *corr;
    sol_t sol={{0}},solf;
    
    nx=NR(opt)+MAXSAT*opt->nf;
    x=zeros(nx,1); P=zeros(nx,nx); xf=mat(NR(opt),1); Qf=mat(NR(opt),NR(opt));
    for (i=0;i<3;i++) sol.rb[i]=opt->rb[i];
    
    /* input observation data and navigation messages for an epoch */
    while ((ny=input(&obs,&nav,&corr))>0) {
        
        /* count rover/reference station observations */
        for (nu=0;nu<ny&&obs[nu].rcv==1;nu++) ;
        for (nr=0;nu+nr<ny&&obs[nu+nr].rcv==2;nr++) ;
        if (opt->mode>=1&&(nu==0||nr==0)) continue;
        
        /* rover position by single point positioning */
        if ((ns=pntpos(obs,nu,nav,opt->elmin,sol.rr,sol.Qr,dtr))>0) stat=5;
        
        if (ns>0&&norm(opt->rb,3)<=0.0) { /* moving baseline */
            pntpos(obs+nu,nr,nav,opt->elmin,sol.rb,Qb,dtr+1);
        }
        if (ns>0&&opt->mode>=1) {
            if (opt->mode<3||norm(x,3)<=0.0) { /* kinematic mode or first epoch? */
                for (i=0;i<3;i++) {
                    x[i]=sol.rr[i];
                    for (j=i;j<nx;j++) P[i+j*nx]=P[j+i*nx]=j<3?sol.Qr[i+j*3]:0.0;
                }
            }
            /* rover position by relative positioning */
            if (!relpos(x,P,nx,obs,nu,nr,nav,corr,&ambc,sol.rb,opt)) {
                
                for (i=0;i<3;i++) {
                    sol.rr[i]=x[i]; for (j=0;j<3;j++) sol.Qr[i+j*3]=P[i+j*nx];
                }
                solf=sol; stat=opt->mode<=1?4:2; /* 1:fix,2:float,4:dgps,5:single */
                
                /* resolve integer ambiguity */
                if (opt->mode>=2&&opt->thresar>=1.0) {
                    if (!resamb(x,P,nx,opt,&ambc,xf,Qf)) {
                        for (i=0,var=0.0;i<3;i++) {
                            solf.rr[i]=xf[i]; var+=Qf[i+i*NR(opt)];
                            for (j=0;j<3;j++) solf.Qr[i+j*3]=Qf[i+j*NR(opt)];
                        }
                        if (sqrt(var)<MAXSTD) stat=1;
                    }
                }
            }
        }
        /* output results */
        if (ns>0) output(timeadd(obs[0].time,-dtr[0]),sol,solf,stat,ns);
    }
    free(x); free(P);
}
